########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import random

from datetime import timedelta

import aioxmpp
import aioxmpp.callbacks
import aioxmpp.errors as errors
import aioxmpp.service as service
import aioxmpp.utils as utils

from . import xso as ibb_xso


MAX_BLOCK_SIZE = (1 << 16) - 1


class IBBTransport(asyncio.Transport):
    """
    The transport for IBB sessions.

    .. note:: Never instantiate this class directly, all instances of
              this class are created the methods
              :meth:`~aioxmpp.ibb.IBBService.open_session` and
              :meth:`~aioxmpp.ibb.IBBService.expect_session` of
              :class:`~aioxmpp.ibb.IBBService`.

    The following keys are supported for
    :meth:`~asyncio.BaseTransport.get_extra_info`:

       `block_size`
           The maximal block size of data in a IBB stanza.

       `peer_jid`
          The JID of the peer.

       `sid`
          The session id of the unerlying IBB session.

       `stanza_type`
          The used stanza type.
    """

    def __init__(self, service, peer_jid, sid, stanza_type, block_size):
        self._protocol = None
        self._service = service
        self._stanza_type = stanza_type
        self._sid = sid
        self._peer_jid = peer_jid
        self._block_size = block_size
        self._incoming_seq = 0
        self._outgoing_seq = 0
        self._closed = False
        self._closing = False

        self.set_write_buffer_limits()
        self._write_buffer = b""
        self._can_write = asyncio.Event()

        self._reading_paused = False
        self._input_buffer = []

        self._write_task = asyncio.ensure_future(self._write_task_main())
        self._write_task.add_done_callback(
            self._handle_close
        )

        self._wait_time = self._service.initial_wait_time.total_seconds()
        self._retries = 0

    def set_write_buffer_limits(self, high=None, low=None):
        if low is None:
            low = 4 * self._block_size

        if high is None:
            high = 8 * self._block_size

        if low < 0:
            raise ValueError("the limits must be positive")

        if high < 0:
            raise ValueError("the limits must be positive")

        if low > high:
            low = high

        self._output_buffer_limit_low = low
        self._output_buffer_limit_high = high

    def get_write_buffer_limits(self):
        return self._output_buffer_limit_low, self._output_buffer_limit_high

    def get_write_buffer_size(self):
        return len(self._write_buffer)

    def set_protocol(self, proto):
        self._protocol = proto
        proto.connection_made(self)

    def get_protocol(self):
        return self._protocol

    def pause_reading(self):
        self._reading_paused = True

    def resume_reading(self):
        self._reading_paused = False
        self._protocol.data_received(b"".join(self._input_buffer))
        self._input_buffer.clear()

    def is_closing(self):
        return self._closing or self._closed

    def get_extra_info(self, key, default=None):
        return {
            "block_size": self._block_size,
            "peer_jid": self._peer_jid,
            "stanza_type": self._stanza_type,
            "sid": self._sid,
        }.get(key, default)

    async def _write_task_main(self):
        e = None
        while True:
            await self._can_write.wait()

            if self._write_buffer:
                data = self._write_buffer[:self._block_size]

                if self._stanza_type == ibb_xso.IBBStanzaType.IQ:
                    stanza = aioxmpp.IQ(
                        aioxmpp.IQType.SET,
                        to=self._peer_jid,
                        payload=ibb_xso.Data(
                            self._sid,
                            self._outgoing_seq,
                            data
                        )
                    )
                elif self._stanza_type == ibb_xso.IBBStanzaType.MESSAGE:
                    # TODO: use some form of tracking for messages
                    stanza = aioxmpp.Message(
                        aioxmpp.MessageType.NORMAL,
                        to=self._peer_jid,
                    )
                    stanza.xep0047_data = ibb_xso.Data(
                        self._sid,
                        self._outgoing_seq,
                        data
                    )

                try:
                    await self._service.client.send(
                        stanza
                    )
                except errors.XMPPWaitError:
                    # wait and try again unless max retries have been reached
                    if self._retries < self._service.max_retries:
                        await asyncio.sleep(self._wait_time)
                        self._wait_time *= self._service.wait_backoff_factor
                        self._retries += 1
                        continue
                    else:
                        e = asyncio.TimeoutError()
                        break
                except errors.StanzaError as _e:
                    # break the loop to close the connection
                    e = _e
                    break
                # update the internal state after the successful
                # write: remove the written data from the buffer and
                # increment the sequence number
                self._write_buffer = self._write_buffer[len(data):]
                self._outgoing_seq += 1
                self._outgoing_seq &= 0xffff

                # reset the wait time
                self._wait_time = \
                    self._service.initial_wait_time.total_seconds()
                self._retries = 0

            if len(self._write_buffer) < self._output_buffer_limit_low:
                self._protocol.resume_writing()

            if not self._write_buffer:
                if self._closing:
                    e = None
                    break
                self._can_write.clear()

        close = ibb_xso.Close()
        close.sid = self._sid
        stanza = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            to=self._peer_jid,
            payload=close,
        )

        try:
            await self._service.client.send(stanza)
        except errors.StanzaError as _e:
            if e is None:
                e = _e
        finally:
            if e is not None:
                raise e

    def write(self, data):
        """
        Send `data` over the IBB. If `data` is larger than the block size
        is is chunked and sent in chunks.

        Chunks from one call of :meth:`write` will always be sent in
        series.
        """

        if self.is_closing():
            return

        self._write_buffer += data

        if len(self._write_buffer) >= self._output_buffer_limit_high:
            self._protocol.pause_writing()

        if self._write_buffer:
            self._can_write.set()

    def _connection_closed(self):
        self._write_task.cancel()

    def _handle_close(self, fut):
        e = None
        self._service._remove_session(self._peer_jid, self._sid)
        try:
            e = fut.exception()
        except asyncio.CancelledError:
            pass
        self._protocol.connection_lost(e)
        self._closed = True

    def close(self):
        """
        Close the session.
        """
        if self.is_closing():
            return

        self._closing = True
        # make sure the writer wakes up
        self._can_write.set()

    def abort(self):
        """
        Abort the session.
        """
        if self.is_closing():
            return
        self._connection_closed()

    def _data_received(self, data):
        if self._closed:
            return

        if self._reading_paused:
            self._input_buffer.append(data)
        else:
            self._protocol.data_received(data)

    def _process_iq(self, payload):
        if payload.seq != self._incoming_seq:
            raise errors.XMPPCancelError(
                condition=errors.ErrorCondition.UNEXPECTED_REQUEST
            )
        self._incoming_seq += 1
        self._incoming_seq &= 0xffff
        self._data_received(payload.content)

    def _process_msg(self, payload):
        if payload.seq != self._incoming_seq:
            return
        self._incoming_seq += 1
        self._incoming_seq &= 0xffff
        self._data_received(payload.content)


class IBBService(service.Service):
    """
    A service implementing in-band bytestreams.

    Methods for establishing sessions:

    .. automethod:: expect_session

    .. automethod:: open_session

    The following attributes control the establishment of sessions due
    to a received request, that was not announced to the service by
    :meth:`expect_session`:

    .. attribute:: session_limit
       :annotation: = 0

       The maximal number of sessions to be accepted. If there are
       that many or more active sessions, no new sessions are
       accepted, unless they are whitelisted by
       :meth:`expect_session`. (This means, that by default only
       expected sessions are accepted!).

    .. attribute:: default_protocol_factory

       The protocol factory to be used when an unexpected connection
       is established. This *must* be set when changing
       :attr:`session_limit` to a non-zero value.

    .. signal:: on_session_accepted(transport, protocol)

       Fires when a session is established due to a received open
       request that was not expected (compare
       :meth:`expect_session`). This can only happen when
       :attr:`session_limit` is set to another value than its default
       value.

    The following attributes control how the IBB sessions react to
    errors of type wait:

    .. attribute:: max_retries
       :annotation: = 5

       The number of times it is tried to resend a data stanza, when a
       :class:`~aioxmpp.errors.XMPPWaitError` is received. When
       :attr:`max_retries` have been tried, the session is closed.
       `connection_lost` of the protocol receives an
       :class:`asyncio.TimeoutError`.

    .. attribute:: initial_wait_time
       :annotation: = timedelta(seconds=1)

       The time to wait when receiving a
       :class:`~aioxmpp.errors.XMPPWaitError` for the first time.

    .. attribute:: wait_backoff_factor
       :annotation: = 1.2

       The factor by which the wait time is prolonged on each
       successive wait error.
    """

    on_session_accepted = aioxmpp.callbacks.Signal()

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)

        self._sessions = {}

        self.session_limit = 0
        self._expected_sessions = {}

        self.default_protocol_factory = None

        self.client.on_stream_destroyed.connect(
            self._on_stream_destroyed
        )

        self.max_retries = 5
        self.initial_wait_time = timedelta(seconds=1)
        self.wait_backoff_factor = 1.2

    def _on_stream_destroyed(self):
        self._expected_sessions = {}

        # tear down the remaining open sessions
        for session in list(self._sessions.values()):
            session.abort()

    def expect_session(self, protocol_factory, peer_jid, sid):
        """
        Whitelist the session with `peer_jid` and the session id `sid` and
        return it when it is established. This is meant to be used
        with signalling protocols like Jingle and is the counterpart
        to :meth:`open_session`.

        :returns: an awaitable object, whose result is the tuple
                  `(transport, protocol)`
        """
        def on_done(fut):
            del self._expected_sessions[sid, peer_jid]

        _, fut = self._expected_sessions[sid, peer_jid] = (
            protocol_factory, asyncio.Future()
        )
        fut.add_done_callback(on_done)
        return fut

    async def open_session(self, protocol_factory, peer_jid, *,
                           stanza_type=ibb_xso.IBBStanzaType.IQ,
                           block_size=4096, sid=None):
        """
        Establish an in-band bytestream session with `peer_jid` and
        return the transport and protocol.

        :param protocol_factory: the protocol factory
        :type protocol_factory: a nullary callable returning an
                                :class:`asyncio.Protocol` instance
        :param peer_jid: the JID with which to establish the byte-stream.
        :type peer_jid: :class:`aioxmpp.JID`
        :param stanza_type: the stanza type to use
        :type stanza_type: class:`~aioxmpp.ibb.IBBStanzaType`
        :param block_size: the maximal size of blocks to transfer
        :type block_size: :class:`int`
        :param sid: the session id to use
        :type sid: :class:`str` (must be a valid NMTOKEN)

        :returns: the transport and protocol
        :rtype: a tuple of :class:`aioxmpp.ibb.service.IBBTransport`
                and :class:`asyncio.Protocol`
        """
        if block_size > MAX_BLOCK_SIZE:
            raise ValueError("block_size too large")

        if sid is None:
            sid = utils.to_nmtoken(random.getrandbits(8*8))
        open_ = ibb_xso.Open()
        open_.stanza = stanza_type
        open_.sid = sid
        open_.block_size = block_size

        # XXX: retry on XMPPModifyError with RESOURCE_CONSTRAINT
        await self.client.send(
            aioxmpp.IQ(
                aioxmpp.IQType.SET,
                to=peer_jid,
                payload=open_,
            )
        )

        handle = self._sessions[sid, peer_jid] = IBBTransport(
            self,
            peer_jid,
            sid,
            stanza_type,
            block_size,
        )

        protocol = protocol_factory()
        handle.set_protocol(protocol)
        return handle, protocol

    @service.iq_handler(
        aioxmpp.IQType.SET,
        ibb_xso.Open)
    async def _handle_open_request(self, iq):
        peer_jid = iq.from_
        sid = iq.payload.sid
        block_size = iq.payload.block_size
        stanza_type = iq.payload.stanza

        if block_size > MAX_BLOCK_SIZE:
            raise errors.XMPPModifyError(
                condition=errors.ErrorCondition.RESOURCE_CONSTRAINT
            )

        try:
            protocol_factory, expected_future = \
                self._expected_sessions[sid, peer_jid]
        except KeyError:
            if len(self._sessions) >= self.session_limit:
                raise errors.XMPPCancelError(
                    condition=errors.ErrorCondition.NOT_ACCEPTABLE
                )
            expected_future = None
            protocol_factory = self.default_protocol_factory

        if (sid, peer_jid) in self._sessions:
            # disallow opening a session twice
            if expected_future is not None:
                # is this correct?
                expected_future.cancel()
            raise errors.XMPPCancelError(
                condition=errors.ErrorCondition.NOT_ACCEPTABLE
            )

        handle = self._sessions[sid, peer_jid] = IBBTransport(
            self,
            peer_jid,
            sid,
            stanza_type,
            block_size
        )

        protocol = protocol_factory()
        handle.set_protocol(protocol)

        if expected_future is None:
            self.on_session_accepted((handle, protocol))
        else:
            expected_future.set_result((handle, protocol))

    @service.iq_handler(
        aioxmpp.IQType.SET,
        ibb_xso.Close)
    async def _handle_close_request(self, iq):
        peer_jid = iq.from_
        sid = iq.payload.sid

        try:
            session_handle = self._sessions[sid, peer_jid]
        except KeyError:
            raise errors.XMPPCancelError(
                condition=errors.ErrorCondition.ITEM_NOT_FOUND
            )

        session_handle._connection_closed()

    @service.iq_handler(
        aioxmpp.IQType.SET,
        ibb_xso.Data)
    async def _handle_data(self, iq):
        peer_jid = iq.from_
        sid = iq.payload.sid

        try:
            session_handle = self._sessions[sid, peer_jid]
        except KeyError:
            raise errors.XMPPCancelError(
                condition=errors.ErrorCondition.ITEM_NOT_FOUND
            )

        session_handle._process_iq(iq.payload)

    @service.inbound_message_filter
    def _handle_message(self, msg):
        if msg.xep0047_data is None:
            return msg

        payload = msg.xep0047_data
        peer_jid = msg.from_
        sid = payload.sid

        try:
            session = self._sessions[sid, peer_jid]
        except KeyError:
            return None

        session._process_msg(payload)

        return None

    def _remove_session(self, peer_jid, sid):
        del self._sessions[sid, peer_jid]



########################################################################
# File name: entity_info.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import itertools

import aioxmpp.disco
import aioxmpp.forms

from framework import Example, exec_example


class Ping(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "target_entity",
            default=None,
            nargs="*",
            type=jid,
            help="Entities to ping",
        )

    async def run_simple_example(self):
        tasks = [
            asyncio.wait_for(aioxmpp.ping.ping(self.client, target), timeout=60)
            for target in self.args.target_entity
        ]
        for addr, result in zip(self.args.target_entity, await asyncio.gather(*tasks, return_exceptions=True)):
            print(addr, result)


if __name__ == "__main__":
    exec_example(Ping())



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.presence` --- Peer presence bookkeeping
######################################################

This module provides a :class:`.PresenceClient` service to track the presence
of peers, no matter whether they are in the roster or not.

.. versionadded:: 0.4

.. currentmodule:: aioxmpp

.. autoclass:: PresenceClient

.. autoclass:: PresenceServer

.. currentmodule:: aioxmpp.presence

.. class:: Service

   Alias of :class:`.PresenceClient`.

   .. deprecated:: 0.8

      The alias will be removed in 1.0.

"""

from .service import PresenceClient, PresenceServer  # NOQA: F401
Service = PresenceClient  # NOQA



########################################################################
# File name: lmc.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

from ..stanza import Message


namespaces.xep0308_replace = "urn:xmpp:message-correct:0"


class Replace(xso.XSO):
    """
    A marker indicating that the stanza is the correction of another one

    .. attribute:: id_

       The identifier of the stanza to correct.
    """

    TAG = namespaces.xep0308_replace, "replace"

    id_ = xso.Attr(
        "id",
    )


Message.xep0308_replace = xso.Child([Replace])



########################################################################
# File name: test_utils.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import types
import sys
import unittest
import unittest.mock

from datetime import timedelta

import aioxmpp.errors as errors
import aioxmpp.utils as utils

from aioxmpp.testutils import (
    CoroutineMock,
    get_timeout,
    run_coroutine,
    make_listener,
)


class Testnamespaces(unittest.TestCase):
    def test_aioxmpp(self):
        self.assertEqual(
            utils.namespaces.aioxmpp_internal,
            "https://zombofant.net/xmlns/aioxmpp#internal"
        )

    def test_namespaces_is_Namespaces_instance(self):
        self.assertIsInstance(
            utils.namespaces,
            utils.Namespaces
        )


class TestNamespaces(unittest.TestCase):
    def setUp(self):
        self.namespaces = utils.Namespaces()
        self.namespaces.aioxmpp_internal = \
            "https://zombofant.net/xmlns/aioxmpp#internal"

    def tearDown(self):
        del self.namespaces

    def test_aioxmpp(self):
        self.assertEqual(
            self.namespaces.aioxmpp_internal,
            "https://zombofant.net/xmlns/aioxmpp#internal"
        )

    def test_inconsistent_rebinding_errors(self):
        with self.assertRaisesRegex(ValueError,
                                    "^inconsistent namespace redefinition$"):
            self.namespaces.aioxmpp_internal = "fnord"

    def test_allow_consistent_rebinding(self):
        self.namespaces.aioxmpp_internal = \
            "https://zombofant.net/xmlns/aioxmpp#internal"

    def test_deleting_is_prohibited(self):
        with self.assertRaisesRegex(AttributeError,
                                    "^deleting short-hands is prohibited$"):
            del self.namespaces.aioxmpp_internal

    def test_namespace_redefinition_errors(self):
        with self.assertRaisesRegex(
                ValueError,
                "^namespace https://zombofant.net/xmlns/aioxmpp#internal"
                " already defined as aioxmpp_internal$"):
            self.namespaces.aioxmpp_internal_clash = \
                "https://zombofant.net/xmlns/aioxmpp#internal"


class Testbackground_task(unittest.TestCase):
    def setUp(self):
        self.coro = CoroutineMock()
        self.coro.return_value = None
        self.started_coro = self.coro()
        self.logger = unittest.mock.Mock()
        self.cm = utils.background_task(
            self.started_coro,
            self.logger,
        )

    def tearDown(self):
        try:
            self.cm.__exit__(None, None, None)
        except:
            pass
        del self.cm
        del self.logger
        del self.coro

    def test_enter_starts_coroutine(self):
        with unittest.mock.patch("asyncio.ensure_future") as async_:
            self.cm.__enter__()

        async_.assert_called_with(self.started_coro)
        self.assertFalse(async_().cancel.mock_calls)

    def test_exit_cancels_coroutine(self):
        with unittest.mock.patch("asyncio.ensure_future") as async_:
            self.cm.__enter__()
            self.cm.__exit__(None, None, None)

        async_().cancel.assert_called_with()

    def test_exit_with_exc_cancels_coroutine_and_propagates(self):
        try:
            raise ValueError()
        except:
            exc_info = sys.exc_info()

        with unittest.mock.patch("asyncio.ensure_future") as async_:
            self.cm.__enter__()
            result = self.cm.__exit__(*exc_info)

        self.assertFalse(result)
        async_().cancel.assert_called_with()

    async def _long_wrapper(self):
        with self.cm:
            await asyncio.sleep(0.1)

    def test_logs_nothing_when_coroutine_terminates_normally(self):
        run_coroutine(self._long_wrapper())
        self.assertFalse(self.logger.mock_calls)

    def test_logs_error_when_coroutine_raises(self):
        async def failing():
            raise ValueError()

        self.cm = utils.background_task(failing(), self.logger)
        run_coroutine(self._long_wrapper())

        self.logger.error.assert_called_with(
            "background task failed: %r",
            unittest.mock.ANY,
            exc_info=True,
        )

    def test_logs_debug_when_coroutine_cancelled(self):
        async def too_long():
            await asyncio.sleep(10)

        self.cm = utils.background_task(too_long(), self.logger)
        run_coroutine(self._long_wrapper())

        self.logger.debug.assert_called_with(
            "background task terminated by CM exit: %r",
            unittest.mock.ANY,
        )

    def test_logs_info_when_coroutine_returns_value(self):
        async def something():
            return unittest.mock.sentinel.result

        self.cm = utils.background_task(something(), self.logger)
        run_coroutine(self._long_wrapper())

        self.logger.info.assert_called_with(
            "background task (%r) returned a value: %r",
            unittest.mock.ANY,
            unittest.mock.sentinel.result,
        )


class Testmagicmethod(unittest.TestCase):
    def test_invoke_on_class(self):
        m = unittest.mock.Mock()

        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                return m(self_or_cls, *args, **kwargs)

        result = Foo.foo(unittest.mock.sentinel.a1,
                         unittest.mock.sentinel.a2,
                         a=unittest.mock.sentinel.a3)

        m.assert_called_once_with(
            Foo,
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            a=unittest.mock.sentinel.a3,
        )

        self.assertEqual(result, m())

    def test_invoke_on_object(self):
        m = unittest.mock.Mock()

        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                return m(self_or_cls, *args, **kwargs)

        instance = Foo()
        result = instance.foo(unittest.mock.sentinel.a1,
                              unittest.mock.sentinel.a2,
                              a=unittest.mock.sentinel.a3)

        m.assert_called_once_with(
            instance,
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            a=unittest.mock.sentinel.a3,
        )

        self.assertEqual(result, m())

    def test_instance_method_is_instance_method(self):
        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                pass

        self.assertIsInstance(
            Foo().foo,
            types.MethodType
        )

    def test_class_method_is_also_method(self):
        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                pass

        self.assertIsInstance(
            Foo.foo,
            types.MethodType
        )

    def test_magicmethod_can_be_overridden(self):
        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                pass

        o = Foo()
        o.foo = "bar"
        self.assertEqual(o.foo, "bar")


class Testmkdir_exist_ok(unittest.TestCase):
    def test_successful_mkdir(self):
        p = unittest.mock.Mock()
        utils.mkdir_exist_ok(p)
        self.assertSequenceEqual(
            p.mock_calls,
            [
                unittest.mock.call.mkdir(parents=True),
            ]
        )

    def test_mkdir_exists_but_is_directory(self):
        p = unittest.mock.Mock()
        p.is_dir.return_value = True
        p.mkdir.side_effect = FileExistsError()
        utils.mkdir_exist_ok(p)
        self.assertSequenceEqual(
            p.mock_calls,
            [
                unittest.mock.call.mkdir(parents=True),
                unittest.mock.call.is_dir()
            ]
        )

    def test_mkdir_exists_but_is_not_directory(self):
        p = unittest.mock.Mock()
        p.is_dir.return_value = False
        exc = FileExistsError()
        p.mkdir.side_effect = exc
        with self.assertRaises(FileExistsError) as ctx:
            utils.mkdir_exist_ok(p)

        self.assertIs(ctx.exception, exc)

        self.assertSequenceEqual(
            p.mock_calls,
            [
                unittest.mock.call.mkdir(parents=True),
                unittest.mock.call.is_dir()
            ]
        )


class TestLazyTask(unittest.TestCase):
    def setUp(self):
        self.coro = CoroutineMock()

    def test_yield_from_able(self):
        self.coro.return_value = unittest.mock.sentinel.result

        async def user(fut):
            return await fut

        fut = utils.LazyTask(self.coro)

        result = run_coroutine(user(fut))

        self.assertEqual(result, unittest.mock.sentinel.result)

    def test_run_coroutine_able(self):
        self.coro.return_value = unittest.mock.sentinel.result

        fut = utils.LazyTask(self.coro)

        result = run_coroutine(fut)

        self.assertEqual(result, unittest.mock.sentinel.result)

    def test_async_able(self):
        self.coro.return_value = unittest.mock.sentinel.result

        fut = utils.LazyTask(self.coro)

        result = run_coroutine(asyncio.ensure_future(fut))

        self.assertEqual(result, unittest.mock.sentinel.result)

    def test_runs_only_once_even_if_awaited_concurrently(self):
        self.coro.return_value = unittest.mock.sentinel.result

        fut = utils.LazyTask(self.coro)

        result2 = run_coroutine(asyncio.ensure_future(fut))
        result1 = run_coroutine(fut)

        self.assertEqual(result1, result2)
        self.assertEqual(result1, unittest.mock.sentinel.result)

        self.coro.assert_called_once_with()

    def test_add_done_callback_spawns_task(self):
        fut = utils.LazyTask(self.coro)
        cb = unittest.mock.Mock(["__call__"])

        with unittest.mock.patch("asyncio.ensure_future") as async_:
            fut.add_done_callback(cb)
            async_.assert_called_once_with(unittest.mock.ANY)

    def test_add_done_callback_works(self):
        fut = utils.LazyTask(self.coro)
        cb = unittest.mock.Mock(["__call__"])

        fut.add_done_callback(cb)

        run_coroutine(fut)

        cb.assert_called_once_with(fut)

    def test_is_future(self):
        self.assertTrue(issubclass(
            utils.LazyTask,
            asyncio.Future,
        ))

    def test_passes_args(self):
        self.coro.return_value = unittest.mock.sentinel.result

        fut = utils.LazyTask(
            self.coro,
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            unittest.mock.sentinel.a3,
        )

        result = run_coroutine(fut)

        self.assertEqual(result, unittest.mock.sentinel.result)

        self.coro.assert_called_once_with(
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            unittest.mock.sentinel.a3,
        )


class Testgather_reraise_multi(unittest.TestCase):

    def test_with_empty_list(self):
        self.assertEqual(
            run_coroutine(utils.gather_reraise_multi()),
            []
        )

    def test_with_one_successful_task(self):
        async def foo():
            return True

        self.assertEqual(
            run_coroutine(utils.gather_reraise_multi(foo())),
            [True]
        )

    def test_with_one_failing_task(self):
        async def foo():
            raise RuntimeError

        try:
            run_coroutine(utils.gather_reraise_multi(foo()))
        except errors.GatherError as e:
            self.assertIs(type(e.exceptions[0]), RuntimeError)
        else:
            self.fail()


    def test_with_two_successful_tasks(self):
        async def foo():
            return True

        async def bar():
            return False

        self.assertEqual(
            run_coroutine(utils.gather_reraise_multi(foo(), bar())),
            [True, False]
        )

    def test_with_two_tasks_one_failing(self):
        async def foo():
            raise RuntimeError

        async def bar():
            return False

        try:
            run_coroutine(utils.gather_reraise_multi(foo(), bar()))
        except errors.GatherError as e:
            self.assertIs(type(e.exceptions[0]), RuntimeError)
        else:
            self.fail()

    def test_with_two_tasks_both_failing(self):
        async def foo():
            raise RuntimeError

        async def bar():
            raise Exception

        try:
            run_coroutine(utils.gather_reraise_multi(foo(), bar()))
        except errors.GatherError as e:
            self.assertIs(type(e.exceptions[0]), RuntimeError)
            self.assertIs(type(e.exceptions[1]), Exception)
        else:
            self.fail()


class Test_to_nmtoken(unittest.TestCase):

    def test_unique_integers(self):
        results = set()
        for i in range(2048):
            res = utils.to_nmtoken(i)
            if res in results:
                self.fail("generated tokens not unique")
            results.add(res)

    def test_integers_and_bytes_do_not_collide(self):
        a = utils.to_nmtoken(255)
        b = utils.to_nmtoken(b"\xff")
        self.assertNotEqual(a, b)
        self.assertTrue(a.startswith(":"))
        self.assertFalse(b.startswith(":"))

    def test_zero_extension(self):
        a = utils.to_nmtoken(b"\xff\x00")
        b = utils.to_nmtoken(b"\xff")
        self.assertNotEqual(a, b)


class TestAlivenessMonitor(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.get_event_loop()
        self.am = utils.AlivenessMonitor(self.loop)
        self.listener = make_listener(self.am)

    def tearDown(self):
        del self.am
        del self.loop

    def test_defaults(self):
        self.assertIsNone(self.am.deadtime_soft_limit)
        self.assertIsNone(self.am.deadtime_hard_limit)

    def test_soft_limit_settable(self):
        self.am.deadtime_soft_limit = timedelta(seconds=1)

        self.assertEqual(self.am.deadtime_soft_limit, timedelta(seconds=1))

    def test_hard_limit_is_settable(self):
        self.am.deadtime_hard_limit = timedelta(seconds=1)

        self.assertEqual(self.am.deadtime_hard_limit, timedelta(seconds=1))

    def test_soft_limit_trips(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_soft_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()

        run_coroutine(asyncio.sleep((dt * 0.2).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()

    def test_notify_received_resets_soft_limit_timer(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_soft_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()

    def test_changing_soft_limit_recaluclates_timer(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_soft_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()

        self.am.deadtime_soft_limit = dt*1.5

        run_coroutine(asyncio.sleep((dt * 0.5).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()

        run_coroutine(asyncio.sleep((dt * 0.2).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()

    def test_soft_timer_fires_even_without_any_reception(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_soft_limit = dt

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()

        run_coroutine(asyncio.sleep((dt * 0.2).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()

    def test_hard_limit_trips(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_hard_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()

        run_coroutine(asyncio.sleep((dt * 0.2).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_called_once_with()

    def test_notify_received_resets_hard_limit_timer(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_hard_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()

    def test_changing_hard_limit_recaluclates_timer(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_hard_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()

        self.am.deadtime_hard_limit = dt*1.5

        run_coroutine(asyncio.sleep((dt * 0.5).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()

        run_coroutine(asyncio.sleep((dt * 0.2).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_called_once_with()

    def test_hard_timer_fires_even_without_any_reception(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_hard_limit = dt

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()

        run_coroutine(asyncio.sleep((dt * 0.2).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_called_once_with()

    def test_both_trip(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_soft_limit = dt
        self.am.deadtime_hard_limit = dt * 1.5
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt * 0.9).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()
        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()

        run_coroutine(asyncio.sleep((dt * 0.2).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()
        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()

        run_coroutine(asyncio.sleep((dt * 0.4).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()
        self.listener.on_deadtime_hard_limit_tripped.assert_called_once_with()

    def test_changing_soft_limit_may_emit_limit_immediately(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt*1.1).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()
        self.am.deadtime_soft_limit = dt

        run_coroutine(asyncio.sleep(0))
        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()

    def test_changing_soft_limit_does_not_reemit_hard_limit(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_hard_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt*1.1).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()
        self.listener.on_deadtime_hard_limit_tripped.assert_called_once_with()

        self.am.deadtime_soft_limit = dt * 1.5

        run_coroutine(asyncio.sleep(0))
        self.listener.on_deadtime_soft_limit_tripped.assert_not_called()
        self.listener.on_deadtime_hard_limit_tripped.assert_called_once_with()

    def test_changing_hard_limit_does_not_reemit_soft_limit(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_soft_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt*1.1).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()
        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()

        self.am.deadtime_hard_limit = dt * 1.5

        run_coroutine(asyncio.sleep(0))
        self.listener.on_deadtime_hard_limit_tripped.assert_not_called()
        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()

    def test_soft_limit_can_reemit_after_reception(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_soft_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt*1.1).total_seconds()))

        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()
        self.am.notify_received()
        self.listener.on_deadtime_soft_limit_tripped.reset_mock()

        run_coroutine(asyncio.sleep((dt*1.1).total_seconds()))
        self.listener.on_deadtime_soft_limit_tripped.assert_called_once_with()

    def test_hard_limit_can_reemit_after_reception(self):
        dt = get_timeout(timedelta(seconds=0.1))

        self.am.deadtime_hard_limit = dt
        self.am.notify_received()

        run_coroutine(asyncio.sleep((dt*1.1).total_seconds()))

        self.listener.on_deadtime_hard_limit_tripped.assert_called_once_with()
        self.am.notify_received()
        self.listener.on_deadtime_hard_limit_tripped.reset_mock()

        run_coroutine(asyncio.sleep((dt*1.1).total_seconds()))
        self.listener.on_deadtime_hard_limit_tripped.assert_called_once_with()


class Testproxy_property(unittest.TestCase):
    def setUp(self):
        self.obj = unittest.mock.Mock(["member"])
        self.obj.member = unittest.mock.Mock(["attr"])
        self.pp = utils.proxy_property("member", "attr")

    def test_forwards_reads(self):
        self.assertEqual(
            self.pp.__get__(self.obj, unittest.mock.sentinel.type_),
            self.obj.member.attr
        )

    def test_forwards_writes(self):
        self.obj.member.attr = unittest.mock.sentinel.old_value

        self.pp.__set__(self.obj, unittest.mock.sentinel.value)

        self.assertEqual(
            self.obj.member.attr,
            unittest.mock.sentinel.value,
        )

    def test_rejects_deletes_by_default(self):
        self.assertTrue(hasattr(self.obj.member, "attr"))

        with self.assertRaisesRegex(AttributeError, "can't delete attribute"):
            self.pp.__delete__(self.obj)

        self.assertTrue(hasattr(self.obj.member, "attr"))

    def test_rejects_writes_if_readonly(self):
        self.obj.member.attr = unittest.mock.sentinel.old_value

        pp = utils.proxy_property("member", "attr", readonly=True)

        with self.assertRaisesRegex(AttributeError, "can't set attribute"):
            pp.__set__(self.obj, unittest.mock.sentinel.value)

        self.assertEqual(
            self.obj.member.attr,
            unittest.mock.sentinel.old_value
        )

    def test_forwards_delete_if_enabled(self):
        self.assertTrue(hasattr(self.obj.member, "attr"))

        pp = utils.proxy_property("member", "attr", allow_delete=True)

        pp.__delete__(self.obj)

        self.assertFalse(hasattr(self.obj.member, "attr"))



########################################################################
# File name: openpgp_legacy.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.stanza
import aioxmpp.xso

from aioxmpp.utils import namespaces

namespaces.xep0027_encrypted = "jabber:x:encrypted"
namespaces.xep0027_signed = "jabber:x:signed"


class OpenPGPEncrypted(aioxmpp.xso.XSO):
    """
    Wrapper around an ASCII-armored OpenPGP encrypted blob.

    .. warning::

        Please see the security considerations of :xep:`27` before making use
        of this protocol. Consider implementation of :xep:`373` instead.

    See :xep:`27` for details.

    .. attribute:: payload

        The character data of the wrapper element.

        .. note::

            While the wire format *is* base64, since the base64 output is
            intended to be passed verbatim to OpenPGP, the payload is declared
            as normal string and aioxmpp will *not* de-base64 it for you (and
            vice versa).
    """

    TAG = namespaces.xep0027_encrypted, "x"

    payload = aioxmpp.xso.Text()


class OpenPGPSigned(aioxmpp.xso.XSO):
    """
    Wrapper around an ASCII-armored OpenPGP signed blob.

    .. warning::

        Please see the security considerations of :xep:`27` before making use
        of this protocol. Consider implementation of :xep:`373` instead.

    See :xep:`27` for details.

    .. attribute:: payload

        The character data of the wrapper element.

        .. note::

            While the wire format *is* base64, since the base64 output is
            intended to be passed verbatim to OpenPGP, the payload is declared
            as normal string and aioxmpp will *not* de-base64 it for you (and
            vice versa).
    """
    TAG = namespaces.xep0027_signed, "x"

    payload = aioxmpp.xso.Text()


aioxmpp.stanza.Message.xep0027_encrypted = aioxmpp.xso.Child([OpenPGPEncrypted])
aioxmpp.stanza.Message.xep0027_signed = aioxmpp.xso.Child([OpenPGPSigned])



########################################################################
# File name: get_external_services.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import itertools

import aioxmpp.extservice
import aioxmpp.forms

from framework import Example, exec_example


class ExternalServices(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "target_entity",
            default=None,
            nargs="?",
            type=jid,
            help="Entity to query (leave empty to query account)"
        )

    async def run_simple_example(self):
        services = await aioxmpp.extservice.get_external_services(
            self.client,
            self.args.target_entity,
        )

        for svc in services.services:
            print("Service:")
            print(f"  type={svc.type_!r}")
            print(f"  transport={svc.transport!r}")
            print(f"  host={svc.host!r}")
            print(f"  port={svc.port!r}")
            print(f"  username={svc.username!r}")
            print(f"  password={svc.password!r}")


if __name__ == "__main__":
    exec_example(ExternalServices())



########################################################################
# File name: utils.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.utils` --- Internal utils
========================================

Miscellaneous utilities used throughout the aioxmpp codebase.

.. data:: namespaces

   Collects all the namespaces from the various standards. Each namespace
   is given a shortname and its value is the namespace string.

   .. note:: This is intended to be promoted to the public API in a
             future release. Third-party users should not assign
             short-names for their own namespaces here, but instead
             use a separate instance of :class:`Namespaces`.

.. autoclass:: Namespaces

.. autofunction:: gather_reraise_multi

.. autofunction:: mkdir_exist_ok

.. autofunction:: to_nmtoken

.. autoclass:: LazyTask

.. autoclass:: AlivenessMonitor

.. autodecorator:: magicmethod

.. autofunction:: proxy_property

"""

import asyncio
import base64
import contextlib
import time
import types

import aioxmpp.callbacks
import aioxmpp.errors

import lxml.etree as etree

__all__ = [
    "etree",
    "namespaces",
]


class Namespaces:
    """
    Manage short-hands for XML namespaces.

    Instances of this class may be used to assign mnemonic short-hands
    to XML namespaces, for example:

    .. code-block:: python

        namespaces = Namespaces()
        namespaces.foo = "urn:example:foo"
        namespaces.bar = "urn:example:bar"

    The class ensures that

    1. only one short-hand is bound to each namespace, so continuing the
       example, the following raises :class:`ValueError`:

       .. code-block:: python

           namespace.example_foo = "urn:example:foo"

    2. no short-hand is redefined to point to a different namespace,
       continuing the example, the following raises
       :class:`ValueError`:

       .. code-block:: python

           namespaces.foo = "urn:example:foo:2"

    3. deleting a short-hand is prohibited, the following raises
       :class:`AttributeError`:

       .. code-block:: python

           del namespaces.foo

    The defined short-hands MUST NOT start with an underscore.

    .. note:: This is intended to be promoted to the public API in a
              future release.
    """

    def __init__(self):
        self._all_namespaces = {}

    def __setattr__(self, attr, value):
        if not attr.startswith("_"):
            try:
                existing_attr = self._all_namespaces[value]
                if attr != existing_attr:
                    raise ValueError(
                        "namespace {} already defined as {}".format(
                            value,
                            existing_attr,
                        )
                    )
            except KeyError:
                try:
                    if getattr(self, attr) != value:
                        raise ValueError("inconsistent namespace redefinition")
                except AttributeError:
                    pass
            self._all_namespaces[value] = attr
        super().__setattr__(attr, value)

    def __delattr__(self, attr):
        if not attr.startswith("_"):
            raise AttributeError("deleting short-hands is prohibited")
        super().__delattr__(attr)


namespaces = Namespaces()
namespaces.xmlstream = "http://etherx.jabber.org/streams"
namespaces.client = "jabber:client"
namespaces.starttls = "urn:ietf:params:xml:ns:xmpp-tls"
namespaces.sasl = "urn:ietf:params:xml:ns:xmpp-sasl"
namespaces.stanzas = "urn:ietf:params:xml:ns:xmpp-stanzas"
namespaces.streams = "urn:ietf:params:xml:ns:xmpp-streams"
namespaces.stream_management = "urn:xmpp:sm:3"
namespaces.aioxmpp = "https://zombofant.net/xmlns/aioxmpp"
namespaces.aioxmpp_test = "https://zombofant.net/xmlns/aioxmpp#test"
namespaces.aioxmpp_internal = "https://zombofant.net/xmlns/aioxmpp#internal"
namespaces.xml = "http://www.w3.org/XML/1998/namespace"


@contextlib.contextmanager
def background_task(coro, logger):
    def log_result(task):
        try:
            result = task.result()
        except asyncio.CancelledError:
            logger.debug("background task terminated by CM exit: %r",
                         task)
        except:  # NOQA
            logger.error("background task failed: %r",
                         task,
                         exc_info=True)
        else:
            if result is not None:
                logger.info("background task (%r) returned a value: %r",
                            task,
                            result)

    task = asyncio.ensure_future(coro)
    task.add_done_callback(log_result)
    try:
        yield
    finally:
        task.cancel()


class magicmethod:
    """
    Decorator for methods that makes them work as instance *and* class
    method.  The first argument will be the class if called on the
    class and the instance when called on the instance.
    """

    __slots__ = ("_f",)

    def __init__(self, f):
        super().__init__()
        self._f = f

    def __get__(self, instance, class_):
        if instance is None:
            return types.MethodType(self._f, class_)
        return types.MethodType(self._f, instance)


def mkdir_exist_ok(path):
    """
    Create a directory (including parents) if it does not exist yet.

    :param path: Path to the directory to create.
    :type path: :class:`pathlib.Path`

    Uses :meth:`pathlib.Path.mkdir`; if the call fails with
    :class:`FileNotFoundError` and `path` refers to a directory, it is treated
    as success.
    """

    try:
        path.mkdir(parents=True)
    except FileExistsError:
        if not path.is_dir():
            raise


class LazyTask(asyncio.Future):
    """
    :class:`asyncio.Future` subclass which spawns a coroutine when it is first
    awaited.

    :param coroutine_function: The coroutine function to invoke.
    :param args: Arguments to pass to `coroutine_function`.

    :class:`LazyTask` objects are awaitable. When the first attempt to await
    them is made, the `coroutine_function` is started with the given `args` and
    the result is awaited. Any further awaits on the :class:`LazyTask` will
    await the same coroutine.
    """

    def __init__(self, coroutine_function, *args):
        super().__init__()
        self.__coroutine_function = coroutine_function
        self.__args = args
        self.__task = None

    def add_done_callback(self, cb, *args):
        self.__start_task()
        return super().add_done_callback(cb, *args)

    def __start_task(self):
        if self.__task is None:
            self.__task = asyncio.ensure_future(
                self.__coroutine_function(*self.__args)
            )
            self.__task.add_done_callback(self.__task_done)

    def __task_done(self, task):
        if task.exception():
            self.set_exception(task.exception())
        else:
            self.set_result(task.result())

    def __iter__(self):
        self.__start_task()
        return iter(self.__task)

    if hasattr(asyncio.Future, "__await__"):
        def __await__(self):
            self.__start_task()
            if hasattr(self.__task, "__await__"):
                return self.__task.__await__()
            else:
                return super().__await__()


async def gather_reraise_multi(*fut_or_coros, message="gather_reraise_multi"):
    """
    Wrap all the arguments `fut_or_coros` in futures with
    :func:`asyncio.ensure_future` and wait until all of them are finish or
    fail.

    :param fut_or_coros: the futures or coroutines to wait for
    :type fut_or_coros: future or coroutine
    :param message: the message included with the raised
        :class:`aioxmpp.errors.GatherError` in the case of failure.
    :type message: :class:`str`
    :returns: the list of the results of the arguments.
    :raises aioxmpp.errors.GatherError: if any of the futures or
        coroutines fail.

    If an exception was raised, reraise all exceptions wrapped in a
    :class:`aioxmpp.errors.GatherError` with the message set to
    `message`.

    .. note::

       This is similar to the standard function
       :func:`asyncio.gather`, but avoids the in-band signalling of
       raised exceptions as return values, by raising exceptions bundled
       as a :class:`aioxmpp.errors.GatherError`.

    .. note::

       Use this function only if you are either

       a) not interested in the return values, or

       b) only interested in the return values if all futures are
          successful.
    """
    todo = [asyncio.ensure_future(fut_or_coro) for fut_or_coro in fut_or_coros]
    if not todo:
        return []

    await asyncio.wait(todo)
    results = []
    exceptions = []
    for fut in todo:
        if fut.exception() is not None:
            exceptions.append(fut.exception())
        else:
            results.append(fut.result())
    if exceptions:
        raise aioxmpp.errors.GatherError(message, exceptions)
    return results


def to_nmtoken(rand_token):
    """
    Convert a (random) token given as raw :class:`bytes` or
    :class:`int` to a valid NMTOKEN
    <https://www.w3.org/TR/xml/#NT-Nmtoken>.

    The encoding as a valid nmtoken is injective, ensuring that two
    different inputs cannot yield the same token. Nevertheless, it is
    recommended to only use one kind of inputs (integers or bytes of a
    consistent length) in one context.
    """

    if isinstance(rand_token, int):
        rand_token = rand_token.to_bytes(
            (rand_token.bit_length() + 7) // 8,
            "little"
        )
        e = base64.urlsafe_b64encode(rand_token).rstrip(b"=").decode("ascii")
        return ":" + e

    if isinstance(rand_token, bytes):
        e = base64.urlsafe_b64encode(rand_token).rstrip(b"=").decode("ascii")
        if not e:
            e = "."
        return  e

    raise TypeError("rand_token musst be a bytes or int instance")


class AlivenessMonitor:
    """
    Monitors aliveness of a data stream.

    .. versionadded:: 0.10

    :param loop: The event loop to operate the checks in.
    :type loop: :class:`asyncio.BaseEventLoop`

    This class is a utility class to implement a traffic-efficient timeout
    mechanism.

    This class can be used to monitor a stream if it is possible to ask the
    remote party send some data (classic ping mechanism). It works particularly
    well if the remote party will send data even without being specifically
    asked for it (saves traffic).

    It is notabily not a mean to enforce a maximum acceptable round-trip time.
    Quite on the contrary, this class was designed specifically to provide a
    reliable experience even *without* an upper bound on the round-trip time.

    To use this class, the using code has to configure the
    :attr:`deadtime_soft_limit`, :attr:`deadtime_hard_limit`, subscribe to the
    signals below and call :meth:`notify_received` whenever *any* data is
    received from the peer.

    There exist two timers, the *soft limit timer* and the *hard limit timer*
    (configured by the respective limit attributes). When the class is
    instantiated, the timers are reset to zero (*but* they start running
    immediately!).

    When a timer exceeds its respective limit, its corresponding signal is
    emitted. The signal is not re-emitted until the next call of
    :meth:`notfiy_received`.

    When :meth:`notify_received` is called, the timers are reset to zero.

    This allows for the following features:

    - Keep a stream alive on a high-latency link as long as data is pouring in.

      This is very useful on saturated (mobile) links. Imagine firing a MAM
      query and a bunch of avatar requests after connecting. With naive ping
      logic, this would easily cause the stream to be considered dead because
      the round-trip time is extremely high.

      However, with this logic, as long as data is pouring in, the stream is
      considered alive.

    - When using the soft limit to trigger a ping and a reasonable difference
      between the soft and the hard limit timeout, this logic gracefully
      reverts to classic pinging when no traffic is seen on the stream.

    - If the peer is pinging us in an interval which works for us (i.e. is less
      than the soft limit), we don’t need to ping the peer; no extra logic
      required.

    This mechanism is used by :class:`aioxmpp.protocol.XMLStream`.

    .. signal:: on_deadtime_soft_limit_tripped()

        Emits when the :attr:`deadtime_soft_limit` expires.

    .. signal:: on_deadtime_hard_limit_tripped()

        Emits when the :attr:`deadtime_hard_limit` expires.

    .. automethod:: notify_received

    .. autoattribute:: deadtime_soft_limit
        :annotation: = None

    .. autoattribute:: deadtime_hard_limit
        :annotation: = None

    """

    on_deadtime_soft_limit_tripped = aioxmpp.callbacks.Signal()
    on_deadtime_hard_limit_tripped = aioxmpp.callbacks.Signal()

    def __init__(self, loop):
        super().__init__()
        self._loop = loop
        self._soft_limit = None
        self._soft_limit_timer = None
        self._soft_limit_tripped = False
        self._hard_limit = None
        self._hard_limit_timer = None
        self._hard_limit_tripped = False
        self._reset_trips()

    def _trip_soft_limit(self):
        if self._soft_limit_tripped:
            return
        self._soft_limit_tripped = True
        self.on_deadtime_soft_limit_tripped()

    def _trip_hard_limit(self):
        if self._hard_limit_tripped:
            return
        self._hard_limit_tripped = True
        self.on_deadtime_hard_limit_tripped()

    def _retrigger_timers(self):
        now = time.monotonic()

        if self._soft_limit_timer is not None:
            self._soft_limit_timer.cancel()
            self._soft_limit_timer = None

        if self._soft_limit is not None:
            self._soft_limit_timer = self._loop.call_later(
                self._soft_limit.total_seconds() - (now - self._last_rx),
                self._trip_soft_limit
            )

        if self._hard_limit_timer is not None:
            self._hard_limit_timer.cancel()
            self._hard_limit_timer = None

        if self._hard_limit is not None:
            self._hard_limit_timer = self._loop.call_later(
                self._hard_limit.total_seconds() - (now - self._last_rx),
                self._trip_hard_limit
            )

    def _reset_trips(self):
        self._soft_limit_tripped = False
        self._hard_limit_tripped = False
        self._last_rx = time.monotonic()

    def notify_received(self):
        """
        Inform the aliveness check that something was received.

        Resets the internal soft/hard limit timers.
        """
        self._reset_trips()
        self._retrigger_timers()

    @property
    def deadtime_soft_limit(self):
        """
        Soft limit for the timespan in which no data is received in the stream.

        When the last data reception was longer than this limit ago,
        :meth:`on_deadtime_soft_limit_tripped` emits once.

        Changing this makes the monitor re-check its limits immediately. Setting
        this to :data:`None` disables the soft limit check.

        Note that setting this to a value greater than
        :attr:`deadtime_hard_limit` means that the hard limit will fire first.
        """
        return self._soft_limit

    @deadtime_soft_limit.setter
    def deadtime_soft_limit(self, value):
        if self._soft_limit_timer is not None:
            self._soft_limit_timer.cancel()
        self._soft_limit = value
        self._retrigger_timers()

    @property
    def deadtime_hard_limit(self):
        """
        Hard limit for the timespan in which no data is received in the stream.

        When the last data reception was longer than this limit ago,
        :meth:`on_deadtime_hard_limit_tripped` emits once.

        Changing this makes the monitor re-check its limits immediately. Setting
        this to :data:`None` disables the hard limit check.

        Note that setting this to a value less than
        :attr:`deadtime_soft_limit` means that the hard limit will fire first.
        """
        return self._hard_limit

    @deadtime_hard_limit.setter
    def deadtime_hard_limit(self, value):
        if self._hard_limit_timer is not None:
            self._hard_limit_timer.cancel()
        self._hard_limit = value
        self._retrigger_timers()


def proxy_property(owner_attr, member_attr, *,
                   readonly=False,
                   allow_delete=False):
    """
    Proxy a property of a member.

    :param owner_attr: The name of the attribute at which the member can
        be found.
    :type owner_attr: :class:`str`
    :param member_attr: The name of the member property to proxy.
    :type member_attr: :class:`str`
    :param readonly: If true, the proxied property will not be writable, even
        if the original property is writable.
    :type readonly: :class:`bool`
    :param allow_delete: If true, the ``del`` operator is allowed on the
        proxy property and will be forwarded to the target property.
    :type allow_delete: :class:`bool`

    .. versionadded:: 0.11.0

    This can be useful when combining classes via composition instead of
    inheritance.

    It is not necessary to set `readonly` to true if the target property is
    already readonly.
    """
    ga = getattr
    sa = setattr
    da = delattr

    def getter(instance):
        return ga(ga(instance, owner_attr), member_attr)

    if readonly:
        setter = None
    else:
        def setter(instance, value):
            return sa(ga(instance, owner_attr), member_attr, value)

    if allow_delete:
        def deleter(instance):
            da(ga(instance, owner_attr), member_attr)
    else:
        deleter = None

    return property(
        getter,
        setter,
        deleter,
    )



########################################################################
# File name: delay.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.xso as xso
from aioxmpp.utils import namespaces

from ..stanza import Message

namespaces.xep0359_stanza_ids = "urn:xmpp:sid:0"


class StanzaID(xso.XSO):
    """
    Represent a :xep:`359` Stanza ID.

    :param id_: The stanza ID to set
    :param by: The entity which has set the stanza ID
    :type by: :class:`aioxmpp.JID`

    .. attribute:: id_

        The assigned stanza ID.

    .. attribute:: by

        The entity who has assigned the stanza ID.

    .. warning::

        Stanza IDs may be spoofed. Please take the security considerations of
        :xep:`359` and the protocols using it into account.
    """
    TAG = (namespaces.xep0359_stanza_ids, "stanza-id")

    id_ = xso.Attr("id", default=None)
    by = xso.Attr("by", type_=xso.JID(), default=None)

    def __init__(self, *, id_=None, by=None, **kwargs):
        super().__init__(**kwargs)
        self.id_ = id_
        self.by = by


class OriginID(xso.XSO):
    """
    Represent a :xep:`359` Origin ID.

    :param id_: The origin ID to set

    .. attribute:: id_

        The assigned origin ID.

    .. warning::

        Origin IDs may be spoofed. Please take the security considerations of
        :xep:`359` and the protocols using it into account.
    """

    TAG = (namespaces.xep0359_stanza_ids, "origin-id")

    id_ = xso.Attr("id", default=None)

    def __init__(self, id_=None):
        super().__init__()
        self.id_ = id_


Message.xep0359_stanza_ids = xso.ChildList([StanzaID])
Message.xep0359_origin_id = xso.Child([OriginID])



########################################################################
# File name: test_delay.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp
import aioxmpp.misc as misc_xso
import aioxmpp.misc.stanzaid as stanzaid_xso
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_namespace(self):
        self.assertEqual(
            namespaces.xep0359_stanza_ids,
            "urn:xmpp:sid:0"
        )


class TestStanzaID(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            misc_xso.StanzaID,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            misc_xso.StanzaID.TAG,
            (namespaces.xep0359_stanza_ids, "stanza-id"),
        )

    def test_default(self):
        si = misc_xso.StanzaID()
        self.assertIsNone(si.id_)
        self.assertIsNone(si.by)

    def test_init(self):
        si = misc_xso.StanzaID(id_="foo", by=aioxmpp.JID.fromstr("bar@baz"))
        self.assertEqual(si.id_, "foo")
        self.assertEqual(si.by, aioxmpp.JID.fromstr("bar@baz"))

    def test_message_attribute(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0359_stanza_ids,
            xso.ChildList,
        )
        self.assertSetEqual(
            aioxmpp.Message.xep0359_stanza_ids._classes,
            {
                stanzaid_xso.StanzaID,
            },
        )


class TestOriginID(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            misc_xso.OriginID,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            misc_xso.OriginID.TAG,
            (namespaces.xep0359_stanza_ids, "origin-id"),
        )

    def test_default(self):
        oi = misc_xso.OriginID()
        self.assertIsNone(oi.id_)

    def test_init(self):
        oi = misc_xso.OriginID(id_="foo")
        self.assertEqual(oi.id_, "foo")

    def test_message_attribute(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0359_origin_id,
            xso.Child,
        )
        self.assertSetEqual(
            aioxmpp.Message.xep0359_origin_id._classes,
            {
                misc_xso.OriginID,
            }
        )



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.ibr` --- In-Band Registration (:xep:`0077`)
##########################################################

This module implements in-band registration.

Registration Functions
======================

The module level registration functions work on an
:class:`aioxmpp.protocol.XMLStream` and before authentication. They
allow to register a new account with a server.

.. autofunction:: get_registration_fields

.. autofunction:: register

Helper Function
===============

.. autofunction:: get_used_fields

Service
=======

.. autoclass:: RegistrationService

XSO Definitions
===============

.. autoclass:: Query
"""

from .service import (  # NOQA: F401
    RegistrationService,
    get_registration_fields,
    register,
)
from .service import get_used_fields  # NOQA: F401
from .xso import Query  # NOQA: F401



########################################################################
# File name: muc_logger.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import configparser
import locale

from datetime import datetime

import aioxmpp.muc
import aioxmpp.structs

from framework import Example, exec_example


class Moderate(aioxmpp.xso.XSO):
    TAG = ("urn:xmpp:message-moderate:0", "moderate")

    retract_flag = aioxmpp.xso.ChildFlag(("urn:xmpp:message-retract:0", "retract"))

    reason = aioxmpp.xso.ChildText(("urn:xmpp:message-moderate:0", "reason"))


@aioxmpp.IQ.as_payload_class
class ApplyTo(aioxmpp.xso.XSO):
    TAG = ("urn:xmpp:fasten:0", "apply-to")

    id_ = aioxmpp.xso.Attr("id")

    payload = aioxmpp.xso.Child([Moderate])


class Moderator(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "--muc",
            type=jid,
            default=None,
            help="JID of the muc to join"
        )

        self.argparse.add_argument(
            "--nick",
            default=None,
            help="Nick name to use"
        )

        self.argparse.add_argument(
            "msgid",
        )

        self.argparse.add_argument(
            "reason",
        )

    def configure(self):
        super().configure()

        self.muc_jid = self.args.muc
        if self.muc_jid is None:
            try:
                self.muc_jid = aioxmpp.JID.fromstr(
                    self.config.get("muc_logger", "muc_jid")
                )
            except (configparser.NoSectionError,
                    configparser.NoOptionError):
                self.muc_jid = aioxmpp.JID.fromstr(
                    input("MUC JID> ")
                )

        self.muc_nick = self.args.nick
        if self.muc_nick is None:
            try:
                self.muc_nick = self.config.get("muc_logger", "nick")
            except (configparser.NoSectionError,
                    configparser.NoOptionError):
                self.muc_nick = input("Nickname> ")

    def make_simple_client(self):
        client = super().make_simple_client()
        muc = client.summon(aioxmpp.MUCClient)
        room, self.room_future = muc.join(
            self.muc_jid,
            self.muc_nick
        )

        return client

    async def run_example(self):
        self.stop_event = self.make_sigint_event()
        await super().run_example()

    async def run_simple_example(self):
        print("waiting to join room...")
        done, pending = await asyncio.wait(
            [
                self.room_future,
                asyncio.create_task(self.stop_event.wait()),
            ],
            return_when=asyncio.FIRST_COMPLETED
        )
        if self.room_future not in done:
            self.room_future.cancel()
            return

        for fut in pending:
            fut.cancel()

        payload = ApplyTo()
        payload.id_ = self.args.msgid
        payload.payload = Moderate()
        payload.payload.reason = self.args.reason
        payload.payload.retract_flag = True

        await self.client.send(aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            to=self.muc_jid,
            payload=payload,
        ))


if __name__ == "__main__":
    exec_example(Moderator())



########################################################################
# File name: oob.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.xso as xso

from ..stanza import Message

from aioxmpp.utils import namespaces

namespaces.xep0066_oob_x = "jabber:x:oob"


class OOBExtension(xso.XSO):
    TAG = namespaces.xep0066_oob_x, "x"

    url = xso.ChildText(
        (namespaces.xep0066_oob_x, "url")
    )


Message.xep0066_oob = xso.Child([OOBExtension])
